/******************************************************************************
*									      *
*	File:     hc2.c							      *
*	Author:   Alison Chaiken					      *
*	Created:  Wed Apr 15 13:15:44 EST 1987				      *
*	Contents: This program is the  root of a routine which calculates     *
*	and  plots the critical  field versus temperature curve		      *
*	given   the various  input  parameters  required.   The		      *
*	following is a list of the variables used:			      *
*									      *
*	     Input Parameters to the Program: 				      *
*                  							      *
*	Parameters with physical significance:				      *
*									      *
*	First, pairb: The number signifying which pairbreaker		      *
*	is to be used.							      *
*									      *
*		Which   pairbreaker  you  use   depends on  the		      *
*	geometry  of your  sample  and  its orientation to  the		      *
*	field.								      *
*									      *
*	(1)	Bulk   sample   or  a   thin   film    oriented		      *
*	perpendicular to the field.					      *
*									      *
*	(2)       Thin film oriented parallel to the field.		      *
*									      *
*	(3)	Surface sheath (Hc3)					      *
*									      *
*									      *
*	0)  slope	dHc2/dT evaluated  at  Tc in  gauss/K.		      *
*	(internally converted to Gauss/K) Not used if pairb = 2.              *
*									      *
*	1) bfulde	Spin-orbit scattering parameter used by		      *
*	Fulde.  b-Fulde =  lambda_{spin-orbit} / 1.122919 where		      *
*	1.122919 = 2 / exp (Euler's constant = 0.5772157).		      *
*									      *
*	2) renorm	Renormalization factor for Pauli field.		      *
*									      *
*	3) cfulde	Fulde's C parameter. Used instead of slope for        *
*                       pairb = 2.
*									      *
*	4)  pp0		Additional pairbreaker  constant  with		      *
*	field.								      *
*									      *
*	5)  pp1		Additional pairbreaker proportionality		      *
*	constant linear with field.					      *
*									      *
*	6) tc		Critical  temperature   in  zero  field		      *
*	(Kelvin).							      *
*									      *
*	7) thick        Thickness of the thin film.  Used only for pairb      *
*                       = 2.                                                  *
*	Variables internal  to the program:                       	      *
*									      *
*	1) npts		number of points in the curve			      *
*									      *
*	2) t		reduced temperature = temp(K) / tc		      *
*									      *
*	3)    h		reduced  field  =    Field(Gauss)  /		      *
*	[(slope)*(tc)]							      *
*									      *
*       4) p[]: Vector into which the 8 parameters are initially read.        *
*                                                                             *
*       5) hcalc: value of hc2 calculated with original parameters input      *
*                 to the program                                              *
*       6) chisqr: the reduced  chi-squared, which is                         *
*                  (chi-squared) / (number of points). Return by hcfit.       *
*                                                                             *
*       7) fit[]: array in which the values of hc2 which minimize chi-squared *
*                 are stored.                                                 *
*                                                                             *
*	Note: Some of the following  are  essentially the same as those       *
*	listed above.  Eric has converted to  these because the		      *
*	actual equations which  are used to calculate Hc2  come		      *
*	form Orlando's  thesis,  which  was written  when   the		      *
*	following parameters were most common.				      *
*									      *
*	8) slambda	lambda_{spin-orbit}				      *
*	9) hp	Pauli field * T_c * 0.9795        			      *
*	10) alpha                                                             *
*                                                                             *
*	Subroutines called by this program:				      *
*                                                                             *
*       1) hcfunc.c, which does the actual calculation of hc2                 *
*                                                                             *
*       2) max.c(), which returns absolute value of argument                  *
*                                                                             *
*       3) hcfit.c, which does gradient least-squares fit, and is only called *
*          if <flag> is present at runtime on the command line.               *
*                                                                             *
*       4) fatal and whiteline, error-handling routines                       *
*                                                                             *
*	C-implementation  by   A.  Chaiken.   Based  on  a		      *
*	Fortran IV routine  written by J.E. Tkaczyk  and J.A.X.		      *
*	Alexander   of  the Francis    Bitter  National  Magnet		      *
*	Laboratory between 1983 and 1986.  				      *
*                                                                             *
*	Please acknowledge J.E. Tkaczyk and J.A.X. Alexander as		      *
*	the source of  this  software.  Also, any requests  you		      *
*	receive for a copy of  this program  should be directed		      *
*	to  the address  below.  We wish   to keep track of the		      *
*	distribution in case any errors or potentially valuable		      *
*	changes are  brought to our attention.  Suggestions and		      *
*	comments are welcome.						      *
*									      *
*	John E. Tkaczyk 						      *
*	MIT Room NW14-2114 						      *
*	Cambridge,  MA 02139						      *
*	(617) 253-5557 							      *
*	(also see APS Directory)					      *
*                                                                             *
*	Copyright (c) 1987 Alison Chaiken.            			      *
*	All rights reserved.						      *
*									      *
*	$Log:	hc2.c,v $
 * Revision 1.1  88/09/06  13:57:05  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include <stdio.h>
#include "defines.h"
#include "paramsdef.h"

#define DELTA_RED 1.122919
#define MAGNETON 3.3586658e-5
/*
 * ARRAYSIZE is the maximum number of lines in a data file;  MAXLINE  is  the
 * maximum  number  of  characters  in  a data line 
 */

char           *fgets_com();

main(argc, argv)
	int             argc;
	char           *argv[];
{
	char            linebuf[MAXLINE],
	               *ifile,
	               *pfile,
	               *ofile;

	int             n,
	                i,
	                j,
	                pairb,
	                c,
	                errflg,
	                fitflag,
	                extflag,
	                mode;

	extern int      optind;
	extern char    *optarg;

	double          x,
	                y,
	                p[NUMPARAMS],
	                hp,
	                chisqr,
	                increm,
	                temp,
	                fit,
	                hcfunc(),
	                max(),
	                hcfit();

	struct paramslist params;
	struct arrayset arrays;

	/*
	 * step is a struct which contains values of the calculated hc2 which
	 * correspond to ARRAYSIZE temperatures below Tc, whereas arrays only
	 * contains values of the calculated hc2 corresponding to the data
	 * points.  the values in arrays are needed to calculate chi-squared in
	 * fit. 
	 */

	FILE           *fopen(), *fpi, *fpj;

	ifile = pfile = ofile = NULL;
	errflg = fitflag = extflag = 0;

	while ((c = getopt(argc, argv, "p:i:o:fe")) != EOF)
		switch (c)
		{
		case 'p':
			pfile = optarg;
			fpi = fopen(pfile, "r");
			if (fpi == NULL)
			{
				perror(pfile);
				exit(-1);
			}
			break;
		case 'i':
			ifile = optarg;
			break;
		case 'o':
			ofile = optarg;
			if (freopen(ofile, "w", stdout) <= (FILE *) 0)
			{
				perror(ofile);
				exit(-1);
			}
			break;
		case 'f':
			fitflag++;
			break;
		case 'e':
			extflag++;
			break;
		case '?':	/* silly huh? */
		default:
			errflg++;
			break;
		}
	if (optind != argc)	/* stuff is left over */
		errflg++;
	if (errflg || (pfile == NULL) || (ifile == NULL))
	{
		fatal("hc2: format is `hc2 -p parametersfile -i datafile [-f] [-o logfile] [-e]`");
	}
	setbuffer(stdout, (char *) 0, 0);
	setbuffer(stderr, (char *) 0, 0);
	/* read in the fit parameters from the parameters file "pfile" */
	n = 0.0;
	(void) fgets_com(&linebuf[0], MAXLINE, fpi);
	if (sscanf(&linebuf[0], "%d", &pairb) != 1)
	{
		if (!whiteline(&linebuf[0]))
			fatal("hc2: The first parameter (pairb) must be an integer. Currently is %s\n", &linebuf[0]);

	}
	if ((pairb != 1) && (pairb != 2) && (pairb != 3))
		fatal("hc2: The pairbreaker must be labelled 1, 2, or 3. Right now is %d : STOP", pairb);

	(void) fgets_com(&linebuf[0], MAXLINE, fpi);
	if (sscanf(&linebuf[0], "%d", &mode) != 1)
	{
		if (!whiteline(&linebuf[0]))
			fatal("hc2: The second parameter (mode) must be an integer. Currently is %s\n", &linebuf[0]);
	}
	if ((mode != 1) && (mode != 0) && (mode != -1))
		fatal("hc2: The mode must be labelled 1, 0, or -1. Right now is %d : STOP", mode);

	while (fgets_com(&linebuf[0], MAXLINE, fpi) != NULL)
	{
		if (sscanf(&linebuf[0], "%lf", &x) != 1)
		{
			if (!whiteline(&linebuf[0]))
				fatal("hc2: Each of the parameters must be a single float. Currently parameter %d is %s", n, &linebuf[0]);
		} else
		{
			p[n] = x;
			n++;
		}

	}

	(void) fclose(fpi);

	if (n != NUMPARAMS)
		fatal("hc2: There must be %d input parameters! Right now there are %d: STOP", NUMPARAMS, n);

	/*
	 * initialize the params structure and switch to the parameters used in
	 * Orlando's thesis 
	 */

	params.pairb = pairb;
	params.mode = mode;
	if (pairb == 2)
	{
		/*
		 * slope is in gauss/kelvin; p[3] = cfulde; p[6] = Tc; where
		 * does 0.0126 come from? 
		 */
		p[3] = max(p[3], 1e-5);	/* prevents overflow */
		params.slope = -(1.26e-2 * p[6] * 1000.0) / p[3];
	} else
		/* slope must be negative  and in g/K, not Kg/K */
	if (p[0] > 0.0)
		params.slope = -p[0];
	else
		params.slope = p[0];


	p[1] = max(p[1], 0.0);	/* prevents overflow if  bfulde = p[1] == 0 */
	params.slambda = p[1] * DELTA_RED;

	/*
	 * DELTA_RED = 1.122919 = (2 * Delta_0)/(PI * T_c0); hp * T_c = Pauli
	 * limiting field.  Pauli field is (Delta_0 * renorm)/sqrt(2). p[6] =
	 * Tc; p[2] = renorm; 1.8572907e4 = [sqrt(2) * PI * k]/[MU * g *
	 * exp(Euler constant)]  where g=2 is electronic g-factor; 0.979587 =
	 * (sqrt(2) * PI * PI)/ (8 * exp(Euler constant)) ; 1.896e4 =
	 * 1.8572907e4 / 0.979587; 
	 */

	if ((hp = FACTOR * p[2]) != 0.0)
		params.alpha = -params.slope / hp;

	else
		fatal("hc2: Tried to divide by zero in calulation of alpha in hc2.\nRenorm == 0 causes overflow");
	params.pp0new = (p[4] * 0.5 / PI);
	params.pp1new = (p[5] / (PI * MAGNETON));
	params.tc = p[6];
	params.thick = p[7];

	/*
	 * MAGNETON = 3.3575e-5 = (mu_Bohr)/(2 * k_Boltzmann); p[4] and p[5]
	 * are actually pp0 and pp1 themselves, but these are the only
	 * combinations in which they are ever used 
	 */

	fprintf(stdout, "The initial parameter values are:\n");
	fprintf(stdout, "pairb = %d; mode = %d; slope = %e; bfulde = %#01.06g;\n",
		params.pairb, params.mode, params.slope, p[1]);
	fprintf(stdout, "renorm = %#01.06g; cfulde = %#01.06g; pp0 = %#01.06g; pp1 = %#01.06g;\n",
		p[2], p[3], p[4], p[5]);
	fprintf(stdout, "Tc = %f; slambda = %#01.06g; alpha = %#01.06g; thick = %#01.06g.\n\n",
		params.tc, params.slambda, params.alpha, params.thick);

	fpj = fopen(ifile, "r");
	if (fpj == NULL)
	{
		perror(ifile);
		exit(3);
	}
	fprintf(stdout, "Red. Temp.    Exp. Hc2    Calc.Hc2\n");
	i = 0;
	while (fgets_com(&linebuf[0], MAXLINE, fpj) != NULL)
	{
		if (sscanf(&linebuf[0], "%lf\t%lf", &x, &y) != 2)
		{
			if (!whiteline(&linebuf[0]))
				fatal("hc2: Can't parse this line: %sin data file %s",
				      &linebuf[0], ifile);
		} else
		{
			arrays.t[i] = x;
			/*
			 * convert to  reduced temperature later, in hcfunc.c 
			 */
			arrays.h[i] = y;	/* assumed in gauss */
			/*
			 * convert to reduced field by dividing by -dHc2/dT and
			 * tc 
			 */
			temp = arrays.t[i] / params.tc;
			arrays.fit[i] = hcfunc(&params, temp);
			/* hcfunc returns value in gauss */
			fprintf(stdout, "%#01.06g      %#01.06g     %#01.06g\n",
			(arrays.t[i] / params.tc), arrays.h[i], arrays.fit[i]);
			/*
			 * fit  contains values of hc2 calculated using the
			 * input parameters now; if hcfit is called, it will
			 * then contain fit values of hc2 
			 */
			i++;
		}

		if (i > ARRAYSIZE)
			fatal("hc2: Maximum number of data pairs is %d: STOP", ARRAYSIZE);
	}
	(void) fclose(fpj);
	i--;			/* i is incremented once more after last pair
				 * is read */
	params.npts = i;

	if (fitflag)
		/* if -f option is given, do least-squares search algorithm */
	{

		chisqr = hcfit(&params, &arrays);
		/* this function call puts values in the array fit */
		fprintf(stdout, "hc2: The final reduced chi-squared value is %#01.06g.\n", chisqr);
		fprintf(stdout, "hc2: The number of degrees of freedom in the calculation of chi-squared is % d.\n ", params.nfree);
		fprintf(stdout, "The final delta values are:\n");
		fprintf(stdout, "delta-slope = %#01.06g; delta-slambda = %#01.06g; delta-alpha = %#01.06g;\n", arrays.delta[0], arrays.delta[1], arrays.delta[2]);
		fprintf(stdout, "delta-pp0new = %#01.06g; delta-pp1new = %#01.06g; delta-tc = %#01.06g;\n", arrays.delta[3], arrays.delta[4], arrays.delta[5]);
		fprintf(stdout, "delta-thick = %#01.06g.\n", arrays.delta[6]);
		fprintf(stdout, "hc2: The final parameter values are:\n");
		/*
		 * the values in params are changed by hcfit just before it
		 * exits 
		 */
		x = -params.alpha * FACTOR; /* set x = slope */
		fprintf(stdout, "pairb = %d; slope = %e; bfulde = %#01.06g;\n",
			params.pairb, x, p[1]);
		fprintf(stdout, "renorm = %#01.06g; cfulde = %#01.06g; pp0 = %#01.06g; pp1 = %#01.06g;\n", p[2], p[3], p[4], p[5]);
		fprintf(stdout, "Tc = %f; slambda = %#01.06g; alpha = %#01.06g; thick = %#01.06g\n\n", params.tc, params.slambda, params.alpha, params.thick);
		fprintf(stdout, "Red. Temp.      Exp. Hc2      Fit Hc2\n");
		for (j = 0; j <= params.npts; j++)
			fprintf(stdout, "%#01.06g        %#01.06g        %#01.06g\n", (arrays.t[j] / params.tc), arrays.h[j], arrays.fit[j]);

	}
	/*
	 * if -e option is used, calculate data points throughout the whole
	 * temperature range, not just at temperatures where data was taken 
	 */
	if (extflag)
	{
		fprintf(stdout, "\n\nExtended temperature range calculation of Hc2:\n");
		fprintf(stdout, "Red. T        Calc. Hc2\n");
		increm = 1.0 / ARRAYSIZE;
		for (i = 1; i < ARRAYSIZE; i++)
			/*
			 * can't calculate hc2 for T exactly equal to 0 (since
			 * must take ln(temp) in hcfunc) 
			 */
		{
			temp = i * increm;
			/* in reduced temperature units */
			fit = hcfunc(&params, temp);
			fprintf(stdout, "%#01.06g     %#01.06g\n", temp, fit);
		}
	}
	fprintf(stderr, "hc2 is finished.\n");
}
