/******************************************************************************
*									      *
*	File:     hcfit.c						      *
*	Author:   Alison Chaiken					      *
*	Created:  Tue May 26 09:47:15 EDT 1987				      *
*	Contents: This subroutine is  called from  the root  program	      *
*	which calculates the critical field.  hcfit.c fits data		      *
*	to  the   calculation in  hcfunc.c by  the  method of a		      *
*	chi-square gradient-search algorithm. Returns value of the reduced    *
*       chi-squared, which is (chi-squared) / (number of degrees of freedom). *
*									      *
*	C-implementation  by   A.  Chaiken.   Based  on  a		      *
*	Fortran IV routine  written by J.E. Tkaczyk  and J.A.X.		      *
*	Alexander   of  the Francis    Bitter  National  Magnet		      *
*	Laboratory between 1983 and 1986.  				      *
*                                                                             *
*	     Subroutines called:					      *
*									      *
*	hcgrad.c: chi-squared gradient search for data fitting.		      *
*	hchisq.c: calculates chi-squared. Called by hcgrad.c.		      *
*									      *
*	Copyright (c) 1987 Alison Chaiken.				      *
*	All rights reserved.						      *
*									      *
*	$Log:	hcfit.c,v $
 * Revision 1.1  88/09/06  13:57:16  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include   <stdio.h>
#include   "paramsdef.h"
#include   "defines.h"

#define MINCHANGE 0.01

double
hcfit(paramsp, arraysp)

	struct paramslist *paramsp;
	struct arrayset *arraysp;

{
	int             n,
	                j,
	                k,
	                counter;

	double          x,
	                chisqr,
	                cstore,
	                change,
	                max(),
	                hchisq(),
	                hcgrad();

	char            linebuf[MAXLINE];
	char           *fgets_com();
	FILE           *fpi;

	fpi = fopen("deltafile", "r");
	if (fpi == NULL)
	{
		perror("deltafile");
		exit(0);
	}
	n = 0;			/* read in fit-parameter increment values */
	counter = 0;
	paramsp->nfree = paramsp->npts - NUMPARAMS + 1;
	/*
	 * paramsp->nfree = minimum number of degrees of freedom; we need to
	 * add to paramsp->nfree the number of delta[j] which are zero to get
	 * the true number of degrees of freedom. We add (NUMPARAMS - 1) rather
	 * than NUMPARAMS because there can be at most seven deltas, since
	 * cfulde can't be directly varied for pairb = 2, only slope can. 
	 */
	while (fgets_com(&linebuf[0], MAXLINE, fpi) != NULL)
	{
		if (sscanf(&linebuf[0], "%lf", &x) != 1)
		{
			if (!whiteline(&linebuf[0]))
				fatal("hcfit: Each of the delta values must be a single float. Currently delta number %d is %s", n, &linebuf[0]);
		} else
		{
			arraysp->delta[n] = x;
			if (arraysp->delta[n] == 0.0)
				/* variation of parameter n is not allowed */
				counter++;
			n++;
		}

	}
	(void) fclose(fpi);

	n--;
	if (n != (NUMPARAMS - 2))
		fatal("hcfit: There are the wrong number (%d) of parameter increment values in the file deltafile: STOP.", n);
	/*
	 * there are (NUMPARAMS - 1) deltas since for pairb = 2 only slope, not
	 * cfulde, can be varied (cfulde is only used for pairb = 2).  In an
	 * array with (NUMPARAMS - 1) elements, the last element is numbered
	 * (NUMPARAMS - 2) since the first is numbered zero.  
	 */

	paramsp->nfree += counter;
	if (paramsp->nfree <= 0)
		fatal("%d data points are not enough to fit with %d variables,", paramsp->npts, ((NUMPARAMS - 1) - counter));
	/*
	 * cstore = latest chi-squared value; chisqr = latest chi-squared
	 * value; fit = array of calculated H_(c2) values 
	 */

	chisqr = hchisq(paramsp, arraysp);
	fprintf(stdout, "hcfit: Initial value of reduced chi-squared is %#01.06g.\n", chisqr);
	cstore = chisqr;
	change = 1.0e6;		/* initialize so that the loop run at least
				 * once */
	k = 0;
	while ((cstore != 0.0) && (change > MINCHANGE) && (k < MAXLOOP))
	{
		for (j = 0; j < n; j++)
			arraysp->delta[j] = (arraysp->delta[j] * 0.5);
		chisqr = hcgrad(paramsp, arraysp);
		if (chisqr != -1.0)
		{
			fprintf(stdout, "hcfit: %dth try at halving of deltas and chisqr is %#01.06g.\n", k, chisqr);
			change = (cstore - chisqr) / cstore;
			change = max(change, -change);
			fprintf(stdout, "hcfit: Change in chi-squared is %#01.06g percent.\n", (100.0 * change));
		} else
			fprintf(stdout, "Underflow in hcgrad; returning previous values.\n");
		fprintf(stdout, "hcfit: Current parameter values are:\n");
		x = -paramsp->alpha * FACTOR;
		fprintf(stdout, "pairb = %d; slope/renorm = %e; slambda = %#01.06g;\n", paramsp->pairb, x, paramsp->slambda);
		fprintf(stdout, "tc = %#01.06g; pp0new = %#01.06g; pp1new = %#01.06g; thick = %#01.06g\n", paramsp->tc, paramsp->pp0new, paramsp->pp1new, paramsp->thick);
		fprintf(stdout, "alpha = %#01.06g.\n\n", paramsp->alpha);
		fprintf(stdout, "Red. Temp.      Fit Hc2\n");
		for (j = 0; j <= paramsp->npts; j++)
			fprintf(stdout, "%#01.06g        %#01.06g\n", (arraysp->t[j] / paramsp->tc), arraysp->fit[j]);
		if (chisqr == -1.0)
			return (cstore);
		/*
		 * chisqr = -1 if increment has gotten too small, so let's
		 * clean up and  exit 
		 */
		cstore = chisqr;
		fprintf(stdout, "hcfit: iteration number = %d; reduced chi-squared = %#01.06g.\n", k, chisqr);
		k++;
	}
	fprintf(stdout,
		"\n\nChi-squared calculated totally %d time%s in hcfit; max allowable is %d.\n",
		k, plural(k), MAXLOOP);
	fprintf(stdout,
		"Final rate of change of chi-squared is %#01.06g percent.\n\n",
		(100.0 * change));
	fprintf(stdout, "Minimum rate of change to continue halving of deltas is %f percent\n", (100 * MINCHANGE));
	return (chisqr);
}
