/******************************************************************************
*									      *
*	File:     hcfunc.c						      *
*	Author:   Alison Chaiken					      *
*	Created:  Tue Apr 21 15:08:08 EST 1987				      *
*	Contents:        This  subroutine    calculates     the		      *
*	critical field for a given temperature,  t, by using a		      *
*	modified linear interpolation procedure.   This program		      *
*	has grown out of the routine HC2-6/21 written by J.A.X.		      *
*	Alexander.  The actual  equations used to calculate Hc2		      *
*	come from T.P.  Orlando's PhD thesis at Stanford.   The		      *
*	present  version of this program is  based on a routine		      *
*	written  by J.E.  Tkaczyk.  C-language transcription by		      *
*	A. Chaiken.							      *
*                                                                             *
*									      *
*	     Variables internal to this program:			      *
*	hzero	field at T=0 if bfulde=0			      *
*	t	reduced temperature = temp(k)/tc			      *
*	h	reduced field = field/[(dHc2/dt)(tc)]			      *
*	hbar	=4*h/(PI*PI)						      *
*	bisect	maximum h/2						      *
*	g	g=0 defines the Hc2 curve				      *
*									      *
*	Subroutines called by this program:				      *
*									      *
*	1)	hcfnyp.c: adds pairbreaker to y				      *
*	2)	hcfnf.c:  calculates the value of g			      *
*       3)      interp.c: modified linear interpolation routine               *
*       4)      max.c: absolute value program                                 *
*       5)      fatal.c and whiteline.c: error-handling routines              *
*									      *
*	Subroutines which calls this program:				      *
*									      *
*	1)	hc2.c							      *
*	2)      hcgrad.c          					      *
*                                                                             *
*	Copyright (c) 1987 Alison Chaiken.				      *
*	All rights reserved.						      *
*									      *
*	$Log:	hcfunc.c,v $
 * Revision 1.1  88/09/06  13:57:19  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include   <math.h>
#include   <stdio.h>
#include   "paramsdef.h"
#include   "defines.h"

#define cubed(x) ((x)*(x)*(x))
#define fourth(x) ((x)*(x)*(x)*(x))
#define fifth(x) ((x)*(x)*(x)*(x)*(x))
#define sixth(x) ((x)*(x)*(x)*(x)*(x)*(x))
#define seventh(x) ((x)*(x)*(x)*(x)*(x)*(x)*(x))
#define PB_HCPERP 1
#define PB_HCPAR 2
#define PB_HC3 3
#define HMIN 5e-12
#define MESHSIZE 5e-4
#define FQUANT 3.12361749278e6

double
hcfunc(paramsp, temp)

	struct paramslist *paramsp;
	double          temp;

{
	double          hbar,
	                bisect,
	                h,
	                g,
	                hstore,
	                gstore,
	                hzero,
	                y,
	                max(),
	                hcfnyp(),
	                hcfnf(),
	                interp();

	/*
	 * The critical field is defined by the condition that g(t,h) = 0.  In
	 * the following we first  find the field, h,  at  which the  value  of
	 * g changes sign and  then pass it to  the  modified linear
	 * interpolation. The parameters (gstore, hstore) and (g, h) are the
	 * values of (g, h) on either side of the zero of g. 
	 */

	/*
	 * * Eric's program converts to reduced temperature units here, but we
	 * have passed the data in reduced temperature units to hcfunc in this
	 * version 
	 */
	if (temp <= 0.0)
		fatal("hcfunc: can't calculate Hc2 for T <= 0! Now T/T_c = %f", temp);
	if (temp >= 1.0)
		return (0.0);
	/* if T > T_c then the critical field is zero! */

	/* Choose the initial values of h, bisect, and gstore  */


	if (paramsp->pairb == PB_HCPAR)
	{
		if ((hzero = (paramsp->slope * paramsp->tc * squared(paramsp->thick))) != 0.0)
			hzero = -2.76471234742e-7 / hzero;
		else
			fatal("hcfunc: Attempt to divide by zero in calculation of hzero: STOP\nCheck the parameters to see if thick, tc, or slope is zero.");

		/*
		 * Eric's program does not explain where the value 2.76e-7
		 * comes from 
		 */
		if (hzero >= 0.0)
			hzero = sqrt(hzero);
		else
			fatal("hcfunc: Attempt to take square root of negative number in calculation of hzero.\nCheck the parameters to see if thick or tc is negative.");

	} else
		hzero = 0.692672873757 / sqrt(1.0 + squared(paramsp->alpha));

	/* h(t = 0.0) for slambda = 0 */

	h = hzero * (1.0 - temp);
	h = max(h, (hzero * 0.05));
	/* Convergence is bad if h starts too small. */

	/* initialization of parameters for bisection loop */
	bisect = h * 0.5;
	gstore = 0;
	hstore = 0;

	/* Bisection loop begins here */
	while (h > HMIN && (bisect > h * MESHSIZE))
	{
		hbar = (h * 4.0) / (PI * PI);
		y = hcfnyp(paramsp, hbar);
		y = y + paramsp->pp0new - (paramsp->pp1new * h * paramsp->slope);
		/* fprintf(stdout, "Calling hcfnf from bisection loop\n"); */
		g = log(temp) - hcfnf(paramsp, temp, hbar, y);
		/* if gstore and g have different signs, begin interpolation */
		if ((gstore * g) < 0.0)
		{
			y = interp(paramsp, temp, hstore, gstore, h, g);
			y = y * paramsp->tc * -paramsp->slope;
			return (y);
			/* return value is in gauss */
		}
		gstore = g;
		hstore = h;
		if (g < 0.0)
			h = h + bisect;
		if (g == 0.0)
			return ((h * paramsp->tc * -paramsp->slope));
		/* return  value is in gauss */
		if (g > 0.0)
		{
			h = h - bisect;
			if (h <= 0.0)
				/*
				 * then we have overshot (h such that g(h) =
				 * 0), so we will decrease the interval size 
				 */
			{
				bisect = bisect * 0.5;
				h = h + bisect;
			}
		}
	}
	return ((h * paramsp->tc * -paramsp->slope));
	/* return value is in gauss */
}

double
hcfnyp(paramsp, hbar)		/* called by hcfunc */

	struct paramslist *paramsp;

	double          hbar;

{
	double          holder;

	switch (paramsp->pairb)
	{

	default:
		fatal("hcfnyp: The pairbreaker must be labelled PB_HCPERP, PB_HCPAR, or PB_HC3. Right now is %d : STOP", paramsp->pairb);
		break;		/* this statement actually is not reached */

	case PB_HCPERP:
		holder = hbar * 0.5;
		break;
		/* bulk of thin film is perpendicular to field */

	case PB_HCPAR:
		holder = (-FQUANT * paramsp->slope * paramsp->tc * squared(paramsp->thick) * squared(hbar));
		break;
		/*
		 * FQUANT = 3.12362e6 = some function of the flux quantum; in
		 * gauss-Ang^2; thin film primarily parallel to field 
		 */

	case PB_HC3:
		/*
		 * 0.5901 is derived on page 86 of Type II Superconductivity by
		 * Saint-James, Thomas, and Sarma 
		 */

		holder = hbar * 0.5 * 0.5901;
		break;

	}
	/* holder is returned for all three cases */
	return (holder);
}


double
hcfnf(paramsp, temp, hbar, y)	/* called by hcfunc */

	struct paramslist *paramsp;

	double          hbar,
	                temp,
	                y;

{
	double          a,
	                b,
	                fnfrea(),
	                fnfima();

	if ((b = (hbar * paramsp->alpha)) != 0.0)
	{
		/*
		 * a is proportional to (bfulde * (1 + squared(alpha)))/(h *
		 * alpha * (1 - t)) 
		 */
		a = paramsp->slambda * 0.5 / b;
		a = max(a, -a);
	} else
		fatal("hcfnf: Attempt to divide by zero.  Either hbar or alpha must equal zero: STOP");

	if (a >= 1.0)
		/* spin--orbit scattering is very important in determining Hc2 */
	{
		/* fprintf(stdout, "Calling fnfrea from hcfnf\n"); */
		b = fnfrea(paramsp, temp, hbar, y);
	}
	if (a < 1.0)
		/*
		 * spin--orbit scattering is NOT very important in determining
		 * Hc2 
		 */
		b = fnfima(paramsp, temp, hbar, y);
	return (b);

}

double
fnfrea(paramsp, temp, hbar, y)	/* called by hcfnf */

	struct paramslist *paramsp;

	double          hbar,
	                temp,
	                y;

{
	double          x,
	                p,
	                q,
	                r,
	                fnpsi();

	x = (squared(paramsp->slambda) * 0.0625) - (squared(hbar) * (squared(paramsp->alpha) * 0.25));
	if (x < 0.0)
		fatal("fnfrea: attempted square root of negative number: STOP");
	x = sqrt(x);
	if (x != 0.0)
		p = paramsp->slambda * 0.25 / x;
	else
		fatal("fnfrea:  Attempted divide by zero (by x): STOP.");

	if (p > 2.0e9)
		p = exp(log(paramsp->slambda / (4.0 * x)));
	q = 0.5 + (y / temp) + ((paramsp->slambda * 0.25) - x) / temp;
	/* grouping is important if paramsp->slambda is large */

	r = 0.5 + (y / temp) + ((paramsp->slambda * 0.25) + x) / temp;
	if ((q < 2.68e5) && (p < 2.68e5))	/* smallest which cause oflow */
		x = fnpsi(0.5) - ((1 + p) * 0.5 * fnpsi(q)) - ((1 - p) * 0.5 * fnpsi(r));
	else
		fatal("fnfrea: p and q are too large for calculation of digamma function\nwithout overflow. Terminated in fnfrea to prevent floating point error\nin fnpsi.  Now p = %e and q = %e.", p, q);
	return (x);		/* not reached for overflow case */
}

#define ERR 0.01

double
fnfima(paramsp, temp, hbar, y)	/* called by hcfnf */

	struct paramslist *paramsp;

	double          hbar,
	                temp,
	                y;

{
	double          ell,
	                fixed,
	                k,
	                alphah,
	                truncerr,
	                bottom,
	                first,
	                chi,
	                angle,
	                factor,
	                second,
	                summ,
	                top;
	int             v,
	                vmax;

	ell = 4.0 * y / (ERR * temp);
	if (ell > 2000.2)
		ell = 2000.2;
	/* use 2000.2 so that vmax will be 5000 at most */
	if (ell < 1.0e-6)
	{
		fprintf(stderr, "fnfima: ell was too small; setting to 1.\n");
		ell = 1.0;
	}
	alphah = paramsp->alpha * hbar;
	k = 2.0 * y * (paramsp->slambda + (2.0 * y)) + squared(alphah);
	/*
	 * calculate truncerr = error in terminating the sum after ell terms,
	 * where ell = 2*v + 1 
	 */

	do
	{
		bottom = ell * ((squared(ell) * (temp) * (temp)) + (temp * ell) * (paramsp->slambda + (4.0 * y)) + k);
		if ((fixed = (ell * squared(ell) * (temp) * (temp) / bottom)) > 0.0)
			first = (log(fixed)) / 2.0;
		else
			fatal("fnfima: tried to take log of a negative number calculation of first: STOP.");
		if ((fixed = ((4.0 * squared(alphah)) - squared(paramsp->slambda))) > 0.0)
			chi = sqrt(fixed);
		else
			fatal("fnfima: Tried to take sqrt of neg. number (chi = %#01.06g): STOP.", chi);
		/* trig functions have arguments in radians */
		angle = atan(((2.0 * ell * temp) + (4.0 * y) + paramsp->slambda) / chi);
		factor = (PI * 0.5) - angle;
		if (chi != 0.0)
			second = paramsp->slambda * factor / chi;
		else
			fatal("fnfima: tried to divide by zero (= chi): STOP.");
		truncerr = -first + second;
		ell = ell * 5.0;
	} while ((ell < 2000.2) && (truncerr > ERR));

	/* Loop to sum up psi series */
	summ = 0.0;
	vmax = 0.5 * (ell - 1);
	/* vmax is integer part of ((ell - 1) / 2) */
	fixed = 2.0 * y * temp;

	for (v = 0.0; v <= vmax; v++)
	{
		ell = (2.0 * v) + 1.0;
		top = (fixed * ell) + k;
		bottom = ell * ((squared(ell) * (temp) * (temp)) + (temp * ell) * (paramsp->slambda + (4.0 * y)) + k);
		summ = summ + (top / bottom);
	}
	/*
	 * fprintf(stdout, "fnfima: psi-sum has looped %d times and the error
	 * is %e\n", vmax, truncerr); 
	 */

	return (-2.0 * summ);
}

double
fnpsi(y)			/* called by fnfrea; returns value of the
				 * digamma function of y */


	double          y;

{
	double          x,
	                digamma(),
	                fracpart;

	int             index,
	                i;

	if (y <= 0.0)
		fatal("fnpsi: unable to calculate the digamma function of zero or negative y; currently y = %f", y);
	if (y < 1.0)
	{
		x = digamma(1.0 + y) - (1.0 / y);
	}
	if ((y <= 2.0) && (y >= 1.0))
		x = digamma(y);
	if ((y < 50.0) && (y > 2.0))
	{
		index = y;	/* integer part of (y - 1) */
		fracpart = y - index;	/* fractional part of y */
		for (i = 0; i < index; i++)
			/* last included term is (index - 1) */
			x = x + (1 / (i + fracpart));
		x = x + digamma(1 + fracpart);
	} else
	if (y >= 50.0)		/* larger than or equal to 50.0 */
	{

		x = log(y) - (0.5 / y) - (1.0 / (12.0 * squared(y))) + (1.0 / (120.0 * fourth(y)));
		x = x - 1.0 / (252.0 * sixth(y));
	}
	return (x);
}


double
digamma(y)
/*
 * called by fnpsi; returns sum of a series which approximates  the digamma
 * function.  Source: Approximations for the Digamma (Psi) Function by W. T.
 * Moody, 1966. 
 */
	double          y;
{
	double          s,
	                seven,
	                sum;

	/*
	 * Source of coefficients is Moody, Approximations for the Digamma
	 * Function 
	 */
	s = y - 1.0;
	seven = seventh(s);
	sum = (0.64493313 * (s - seven)) - (0.20203181 * (squared(s) - seven)) + (8.209433e-2 * (cubed(s) - seven));
	sum = sum - (3.591665e-2 * (fourth(s) - seven)) + (1.485925e-2 * (fifth(s) - seven));
	sum = sum - (4.72505e-3 * (sixth(s) - seven));
	sum = sum + (s / (1.0 + s)) - 0.5772156649 + (seven * 0.5);
	return (sum);
}

#define XTOL 5e-4
#define YTOL 0.0
#define NLIM 100

double
interp(paramsp, temp, xstore, ystore, x, y)

	struct paramslist *paramsp;

	double          xstore,
	                ystore,
	                x,
	                y,
	                temp;


{
	double          ysave,
	                newx,
	                xerr,
	                yerr,
	                hbar,
	                hcfnf(),
	                hcfnyp();
	int             j;

	ysave = ystore;
	/* initialize the errors to large values */
	yerr = 1.0e6;
	xerr = 1.0e6;
	for (j = 1; j <= NLIM; j++)
	{
		newx = x - (y * (x - xstore) / (y - ystore));
		/* newx = oldx - (oldy / slope)  */
		hbar = newx * 4 / (PI * PI);
		/*
		 * fprintf(stdout, "Calling hcfnf from interpolation
		 * routine\n"); 
		 */
		yerr = hcfnyp(paramsp, hbar);
		yerr = yerr + paramsp->pp0new - (paramsp->pp1new * newx * paramsp->slope);
		yerr = log(temp) - hcfnf(paramsp, temp, hbar, yerr);
		/* yerr = g (newx) */
		xerr = (xstore - x) / x;
		if (max(yerr, -yerr) <= YTOL)
		{
			/*
			 * fprintf(stdout, "Interpolation converged in
			 * y-tolerance in %d iterations.  X-error is
			 * %#01.06g.\n", j, xerr); 
			 */
			return (newx);
		}
		if (max(xerr, -xerr) <= XTOL)
		{
			/*
			 * fprintf(stdout, "Interpolation converged in
			 * x-tolerance in %d iterations.  Y-error is
			 * %#01.06g.\n\n", j, yerr); 
			 */
			return (newx);
		}
		if ((yerr * ystore) < 0.0)
			/*
			 * then the interval (yerr, ystore) still contains g =
			 * 0, so let's just make the interval smaller 
			 */
		{
			x = newx;
			y = yerr;
			if ((yerr * ysave) > 0.0)
				ystore = ystore * 0.5;
		} else
		{
			xstore = newx;
			ystore = yerr;
			if ((yerr * ysave) > 0.0)
				y = y * 0.5;
		}
		ysave = yerr;
	}
	fprintf(stdout, "Maximum number of iterations (%d) exceeded in interpolation routine.\n", NLIM);
	fprintf(stdout, "Returning the final value of x = %#01.06g with xerr = %#01.06g and yerr = %#01.06g.\n\n", newx, xerr, yerr);
	return (newx);
}
