/******************************************************************************
*									      *
*	File:     hcgrad.c						      *
*	Author:   Alison Chaiken					      *
*	Created:  Wed May 27 12:47:39 EDT 1987				      *
*	Contents: Origin: Program  11-3 GRADLS from  P.R. Bevington, Data     *
*	Reduction and Error Analysis for the Physical Sciences.		      *
*	(Has been modified somewhat.)					      *
*                                                                             *
*	C-implementation  by   A.  Chaiken.   Based  on  a		      *
*	Fortran IV routine  written by J.E. Tkaczyk  and J.A.X.		      *
*	Alexander   of  the Francis    Bitter  National  Magnet		      *
*	Laboratory between 1983 and 1986.  				      *
*	     								      *
*	Purpose: Make a  gradient--search least-squares fit  to		      *
*	data  with a specified function  which is  not linear in	      *
*	coefficients.							      *
*									      *
*	Parameters:							      *
*		arrays->t: array of data points for independent variable      *
*        	arrays->y: array of data points for dependent variable	      *
*		arrays->sigmay: array of standard deviations for y variable   *
*		mode:  determines methos of weighting least-squares           *
*                              fit:                                           *
*				+1 (instrumental weight) =                    *
*                                       1/ sigmay(i) * sigmay(i)              *
*				 0 (no weighting) = 1			      *
*				-1 (statistical weighting) = 1/ y(i)	      *
*		delta: array of increments for the input parameters           *
*		arrays->fit: array of calculated parameters for y	      *
*		chisqr: reduced chisqr value			              *
*									      *
*	Subroutines which are called:					      *
*									      *
*		hcfunc.c						      *
*		hchisq.c						      *
*									      *
*	Called by: hc2.c						      *
*									      *
*	Copyright (c) 1987 Alison Chaiken.				      *
*	All rights reserved.						      *
*									      *
*	$Log:	hcgrad.c,v $
 * Revision 1.1  88/09/06  13:57:24  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include   <stdio.h>
#include   <math.h>
#include   "paramsdef.h"
#include   "defines.h"

#define GRIDSIZE 0.1
#define MININCREM 1.0e-8

double
hcgrad(paramsp, arraysp)

	struct paramslist *paramsp;
	struct arrayset *arraysp;

{
	struct paramslist temp;
	double          grad[(NUMPARAMS - 1)],
	                chisave,
	                chisqr,
	                redtemp,
	                hcfunc(),
	                hchisq(),
	                increm,
	                sum,
	                newchisq;

	int             i,
	                j,
	                k;

	temp = *paramsp;

	/* calculate chi-squared with the input parameters */
	chisave = hchisq(paramsp, arraysp);

	/*
	 * values of hc2 calculated for the input parameters have already been
	 * put into arraysp->fit[] by the root program hc2 which calls hcgrad 
	 */

	/* evaluate the gradient of chi-squared */
	sum = 0.0;
	/*
	 * there are (NUMPARAMS - 1) deltas since for pairb = 2 only slope, not
	 * cfulde, can be varied (cfulde is only used for pairb = 2).  In an
	 * array with (NUMPARAMS - 1) elements, the last element is numbered
	 * (NUMPARAMS - 2) since the first is numbered zero.  therefore loop
	 * should exit after j = NUMPARAMS - 2. 
	 */
	for (j = 0; j < (NUMPARAMS - 1); j++)
	{
		if (arraysp->delta[j] == 0.0)
		{
			/* variation of parameter j is not allowed */
			grad[j] = 0;
		} else
		{
			increm = (1 + GRIDSIZE) * arraysp->delta[j];
			changeparams(j, &temp, paramsp, increm);
			for (i = 0; i <= paramsp->npts; i++)
			{
				redtemp = arraysp->t[i] / temp.tc;
				arraysp->fit[i] = hcfunc(&temp, redtemp);
			}
			grad[j] = chisave - hchisq(paramsp, arraysp);
			/* must make sure hchisq returns a dimensionless value */
			sum = sum + ((grad[j]) * (grad[j]));
		}
	}
	for (k = 0; k < (NUMPARAMS - 1); k++)
	{
		grad[k] = (arraysp->delta[k] * grad[k]) / sqrt(sum);
		grad[k] = grad[k] * 2.0;
		/*
		 * this is done because grad[k] is divided by 2 at the
		 * beginning of the next loop 
		 */
	}


	chisqr = 1.0e6;
	/*
	 * evaluate chi-squared at the new point now that gradient has been
	 * calculated; initialize chisq so that loop is done at least once 
	 */
	j = 0;
	while ((chisave <= chisqr) && (j < MAXLOOP))	/* as long as
							 * chi-squared keeps
							 * increasing */
	{
		for (k = 0; k < (NUMPARAMS - 1); k++)
		{
			grad[k] = grad[k] * 0.5;
			if ((grad[k] < MININCREM) && (grad[k] > -MININCREM) &&
			    (arraysp->delta[k] != 0.0))
			{
				fprintf(stdout, "hcgrad: grad[%d] is getting too small: %#01.06g.\n", k, grad[k]);
				fprintf(stdout, "hcgrad: Minimum grad[] allowed is %e.\n", MININCREM);
				fprintf(stdout, "hcgrad: Returning and exiting.\n");
				return (-1.0);
			}
			/*
			 * decrease increment since chi-squared has been
			 * increasing 
			 */
			changeparams(k, &temp, paramsp, grad[k]);
			/*
			 * temp is the INPUT value of parameter plus grad;
			 * changes in the parameters are not accumulating 
			 */
		}

		for (i = 0; i <= paramsp->npts; i++){
		     redtemp = arraysp->t[i] /temp.tc;
			arraysp->fit[i] = hcfunc(&temp, redtemp);
		   }
		chisqr = hchisq(paramsp, arraysp);
		j++;
	}

	fprintf(stdout, "hcgrad:  Went through %d iteration%s in order to calculate the gradient of\nchi-squared. Max number is %d.\n",
		j, plural(j), MAXLOOP);

	/*
	 * now chi-squared is finally decreasing; let's increment parameters
	 * with grad[] of the order of magnitude determined by the last loop
	 * until chi-squared begins to increase. Point of previous loop was to
	 * determine appropriate values of grad[]; now we will use these to
	 * minimize chi-squared. 
	 */


	fprintf(stdout, "\nhcgrad: The last three values of chi-squared are:\n");
	/* now we initialize newchisq */
	for (j = 0; j < (NUMPARAMS - 1); j++)
		changeparams(j, &temp, &temp, grad[j]);
	for (k = 0; k <= paramsp->npts; k++){
	     redtemp = arraysp->t[k] / temp.tc;
	     arraysp->fit[k] = hcfunc(&temp, redtemp);
	}
	
	newchisq = hchisq(paramsp, arraysp);
	i = 0;
	do			/* move this way as long as chi-squared is
				 * decreasing */
	{
		chisave = chisqr;
		chisqr = newchisq;
		for (j = 0; j < (NUMPARAMS - 1); j++)
			changeparams(j, &temp, &temp, grad[j]);
		/*
		 * now change in the parameters is accumulating, as both struct
		 * arguments passed to changeparams are temp 
		 */
		for (k = 0; k <= paramsp->npts; k++){
		     redtemp = arraysp->t[k] / temp.tc;
		     arraysp->fit[k] = hcfunc(&temp, redtemp);
		}
		newchisq = hchisq(paramsp, arraysp);
		fprintf(stdout, "%d    %#01.06g    %#01.06g    %#01.06g\n", i, chisave, chisqr, newchisq);
		i++;
	} while ((newchisq < chisqr) && (i < MAXLOOP));

	fprintf(stdout, "hcgrad:  Chi-squared decreased for %d increment%s of the parameters. Maximum\nnumber of iterations is %d.\nFinal chi-squared for this call of hcgrad is chi-squared is %#01.06g.\n\n", i, plural(i), MAXLOOP, chisqr);
	/*
	 * change params to best value, which minimized chisqr.  Note that this
	 * is the first time in the program that the copy of the parameters in
	 * the struct params is changed. 
	 */
	for (j = 0; j < (NUMPARAMS - 1); j++)
		changeparams(j, paramsp, &temp, -grad[j]);
	/*
	 * chi-squared has just begun increasing, so do parabolic interpolation
	 * to find the approximate minimum (exact minimum if chi-squared is
	 * parabolic in functional form near the minimum) 
	 */
	if ((increm = ((chisave - chisqr) * (newchisq - chisqr))) == -1.0)
		fatal("hcgrad: attempted divide by zero in parabolic interpolation routine.\n");
	increm = 0.5 + 1.0 / (1.0 + increm);
	for (j = 0; j < (NUMPARAMS - 1); j++)
		changeparams(j, &temp, paramsp, (-increm * grad[j]));
	for (k = 0; k <= paramsp->npts; k++){
	     redtemp = arraysp->t[k] / temp.tc;
	     arraysp->fit[k] = hcfunc(&temp, redtemp);
	}
	newchisq = hchisq(paramsp, arraysp);
	if (chisqr < newchisq)
	{
		/*
		 * now that we have found the minimum of chi-squared, make sure
		 * that the corresponding values of the parameters and fit are
		 * in the arrrays 
		 */
		increm--;
		/*
		 * Not needed since we changed only temp in calculation of
		 * newchisq: for (j = 0; j < (NUMPARAMS - 1); j++)
		 * changeparams(j, &temp, paramsp, (-increm * grad[j])); 
		 */
		for (k = 0; k <= paramsp->npts; k++){
		     redtemp = arraysp->t[k] / temp.tc;
		     arraysp->fit[k] = hcfunc(paramsp, redtemp);
		}
	} else
	{
		*paramsp = temp;
		chisqr = newchisq;
	}

	return (chisqr);
}

changeparams(j, astruct, bstruct, increm)

	int             j;
	double          increm;

	struct paramslist *astruct,
	               *bstruct;
/*
 * in this subroutine, astruct is modified, but bstruct is not.  astruct is
 * temporary copy of values of parameters whose input values are stored in the
 * struct bstruct.  bstruct is the last value of astruct (originally =
 * paramsp). 
 */

{
	switch (j)
	{
	default:
		fatal("changeparams: No such parameter %d.: STOP", j);
		break;

	case 0:
		astruct->slope = bstruct->slope + increm;
		break;

	case 1:
		astruct->slambda = bstruct->slambda + increm;
		break;

	case 2:
		astruct->alpha = bstruct->alpha + increm;
		break;

	case 3:
		astruct->pp0new = bstruct->pp0new + increm;
		break;

	case 4:
		astruct->pp1new = bstruct->pp1new + increm;
		break;

	case 5:
		astruct->tc = bstruct->tc + increm;
		break;

	case 6:
		astruct->thick = bstruct->thick + increm;
		break;
	}
	return;
}
