/******************************************************************************
*									      *
*	File:     hchisq.c						      *
*	Author:   Alison Chaiken					      *
*	Created:  Sun May 31 10:55:35 EDT 1987				      *
*	Contents:                                                             *
*	        Origin:  P.R. Bevington, Data     			      *
*		Reduction and Error Analysis for the Physical Sciences.	      *
*		(Has been modified somewhat.)				      *
*	                                                                      *
*		C-implementation  by   A.  Chaiken.   Based  on  a	      *
*		Fortran IV routine  written by J.E. Tkaczyk  and J.A.X.	      *
*		Alexander   of  the Francis    Bitter  National  Magnet	      *
*		Laboratory between 1983 and 1986.  			      *
*		     							      *
*	        Purpose: Evaluate the reduced chi-squared for fit to data.    *
*	           hchisq = sum (squared(yfit[] - y[]) /                      *
*                  (squared(sigmay[]) * paramsp->nfree))                      *
*								              *
*		Parameters:                                                   *
*                                                                             *
*	                arrays->y: array of data points for dependent variable*
*			arrays->sigmay: array of standard deviations for y    *
*                                       variable                              *
*			mode:  determines method of weighting least-squares   *
*	                              fit:                                    *
*					+1 (instrumental weight) =            *
*	                                       1/ sigmay(i)  sigmay(i)        *
*					 0 (no weighting) = 1		      *
*					-1 (statistical weighting) = 1/ y(i)  *
*			delta: array of increments for the input parameters   *
*			arrays->fit: array of calculated parameters for y     *
*			chisqr: reduced chisqr value			      *
*	                params->npts: number of data points		      *
*	                paramsp->nfree:  number of degrees of freedom         *
*	     								      *
*	     Subroutines which are called: none				      *
*	     								      *
*	     Called by: hcgrad.c and hcfit.c				      *
*									      *
*	Modified  11/18/87     to calculate    chi-squared   in		      *
*	dimensionless units  (rather than in gauss) by dividing		      *
*	through by the experimental  value  SQUARED rather than		      *
*	linearly.							      *
*                                                                             *
*	Copyright (c) 1987 Alison Chaiken.				      *
*	All rights reserved.						      *
*									      *
*	$Log:	hchisq.c,v $
 * Revision 1.1  88/09/06  13:57:28  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include "paramsdef.h"
#include "defines.h"

double
hchisq(paramsp, arraysp)

	struct paramslist *paramsp;
	struct arrayset *arraysp;

{
	double          chisq,
	                weight,
	                holder,
	                max();

	int             i;

	chisq = 0;
	switch (paramsp->mode)
	{
	default:
		fatal("hchisq: Mode must be 1, 0, or -1; now is %d. STOP.", paramsp->mode);
		break;		/* not reached */

	case 1:
		for (i = 0; i <= paramsp->npts; i++)
		{
			if ((weight = ((arraysp->sigmay[i]) * (arraysp->sigmay[i]))) != 0.0)
			     weight = 1 / weight;
			else
				fatal("hchisq: divide by zero in attempt to calculate weight of chi-squared.  Make sure that none of the sigmay[] are zero. STOP");
			holder = arraysp->h[i] - arraysp->fit[i];
			chisq = chisq + weight * (squared(holder));
		}
		break;

	case 0:
		/* weight = 1.0 */
		for (i = 0; i <= paramsp->npts; i++)
		{
			holder = arraysp->h[i] - arraysp->fit[i];
			chisq = chisq + (squared(holder));

		}
		break;

	case -1:
		for (i = 0; i <= paramsp->npts; i++)
		{
			if (arraysp->h[i] == 0.0)
				weight = 1.0;
			else
			{
				weight = max(arraysp->h[i], -arraysp->h[i]);
				weight = 1.0 / squared(weight);
			}
			holder = arraysp->h[i] - arraysp->fit[i];
			chisq = chisq + (weight * (squared(holder)));
		}
		break;

	}

	chisq = chisq / paramsp->nfree;	/* we are returning the reduced
					 * chi-squared */
	return (chisq);
}
