/******************************************************************************
*									      *
*	File:     hczero.c						      *
*	Author:   Alison Chaiken					      *
*	Created:  Tue Feb 17 14:42:38 EST 1987				      *
*	Contents: Fits a line to  Hc vs.   T  data using  a least-squares     *
*	algorithm.  Algorithm is from Gerald, Applied Numerical		      *
*	Analysis, Second Edition,  p. 468.  Modified  5/5/87 to		      *
*	use fgets_com instead of  fgetline. Modified 6/12/87 to		      *
*	print out  the  reduced  chi-squared,  in order  to  be		      *
*	consistent with output of the program hc2.c.			      *
*									      *
*	Modified 11/18/87 to calculate a reduced chi-squared in		      *
*	dimensionless  units,  rather than  gauss,  by dividing		      *
*	through by the square of the experimental value, rather		      *
*	than its value.							      *
*                                                                             *
*	Subroutines called:						      *
*	fatal.c     error handler					      *
*	whiteline.c  detects blank lines in input			      *
*	fgets_com.c  strips comments from data file, and passes 	      *
*	             lines to fget.c					      *
*                                                                             *
*	$Log:	hczero.c,v $
 * Revision 1.3  88/09/27  15:35:14  alison
 * Modified to calculate the mean error per data point
 * (error-bar length).
 * 
 * Revision 1.2  88/09/26  17:13:18  alison
 * Rewritten to use getopt instead of looking at argv[] directly.
 * Now can take -q flag to do a quadratic fit rather than a
 * linear one.
 * 
 * Revision 1.1  88/09/06  13:57:32  alison
 * Initial revision
 * 								      *
******************************************************************************/

#include <stdio.h>
#include <math.h>

#define SIZE 200
#define MAXLINE 129

char           *fgets_com();

main(argc, argv)
	int             argc;
	char           *argv[];
{
	char            linebuf[MAXLINE],
	               *ifile,
	               *ofile;

	int             n,
	                i,
	                c,
	                errflg,
	                qflag;

	extern int      optind;
	extern char    *optarg;


	double          t[SIZE],
	                h[SIZE],
	                x,
	                y,
	                a,
	                b,
	                tsum,
	                hsum,
	                thsum,
	                ttsum,
	                det,
	                varsq,
	                var,
	                sigmasq,
	                chisq,
	                tceff;

	FILE           *fopen(), *fpi, *fpj;

	qflag = 0;
	ifile = ofile = NULL;
	errflg = 0;

	while ((c = getopt(argc, argv, "i:o:q")) != EOF)
		switch (c)
		{
		case 'q':	/* do a quadratic fit */
			qflag++;
			break;
		case 'i':	/* datafile */
			ifile = optarg;
			fpi = fopen(ifile, "r");
			if (fpi == NULL)
			{
				perror(ifile);
				exit(-1);
			}
			break;
		case 'o':	/* outputfile */
			ofile = optarg;
			fpj = fopen(ofile, "w");

			if (fpj == NULL)
			{
				perror(ofile);
				exit(-1);
			}
			break;
		default:
			errflg++;
			break;
		}
	if (optind != argc)	/* stuff is left over */
		errflg++;
	if (errflg || (ifile == NULL))
	{
		fatal("hczero: format is `hczero -i infile <-o outfile> <-q>.': STOP.\nThe qflag option is for quadratic fits.");
	}
	/* if no output file, put results at end of input file */
	if (ofile == NULL)
	{
		fpj = fopen(ifile, "a");
		if (fpj == NULL)
		{
			perror(ofile);
			exit(-1);
		}
	}
	tsum = 0;
	hsum = 0;
	thsum = 0;
	ttsum = 0;
	n = 0;
	while (fgets_com(&linebuf[0], MAXLINE, fpi) != NULL)
	{
		if ((sscanf(&linebuf[0], "%f\t%f", &x, &y) != 2))
		{
			if (!whiteline(&linebuf[0]))

				fprintf(stderr, "****I can't parse this line\n**** %s\n", &linebuf[0]);
		} else
		{
			if (qflag == 0)
				t[n] = x;
			else
				t[n] = x * x;
			h[n] = y;
			tsum += t[n];
			hsum += h[n];
			thsum += t[n] * h[n];
			ttsum += t[n] * t[n];
			n++;
		}
	}
	fprintf(fpj, "\n# Number of points = %d.\n", n);

	if ((n == 0) || (n > SIZE))
		fatal("Number of points must be between 1 and %d;\n now is %d", SIZE, n);

	if (n == 1)
		fatal("I need more than one data point to do a curve fit!");
	if ((det = (n * ttsum) - (tsum * tsum)) == 0)
		fatal("ALL of these points are the same!");
	det = 1 / det;
	b = ((ttsum * hsum) - (tsum * thsum)) * det;
	a = (n * thsum - hsum * tsum) * det;
	/* h(T) = (a x T)  + b */

	if (qflag == 0)
	{
		fprintf(fpj, "# Hc2 = (%.3f)*T + (%.3f).\n", a, b);
		tceff = -b / a;
	} else
	{
		fprintf(fpj, "# Hc2 = (%.3f)*T^2 + (%.3f).\n", a, b);
		tceff = sqrt(-(b / a));
	}

	fprintf(fpj, "# Effective Tc is %.2f K.\n", tceff);
	sigmasq = 0.0;
	chisq = 0.0;
	for (i = 0; i < n; i++)
	{
		if (h[i] != 0.0)
		{
			var = (h[i] - ((a * t[i]) + b));
			varsq = var * var;
			sigmasq += varsq;
			chisq = chisq + (varsq / (h[i] * h[i]));
		} else
			n--;	/* we're no longer including this point in
				 * chi-squared */
	}
	n = n - 2;
	/*
	 * number of points - 2 = number of degrees of freedom for a line fit 
	 */
	if (n != 0)
		chisq = chisq / n;
	else
		fatal("Not enough data for a two-parameter fit!");
	sigmasq = sqrt(sigmasq / n);
	if (qflag == 0)
		fprintf(fpj, "# Reduced chi-squared for the line fit = %.2e\n# with statistical weighting.\n", chisq);
	else
		fprintf(fpj, "# Reduced chi-squared for the quadratic fit = %.2e\n# with statistical weighting.\n", chisq);
	fprintf(fpj, "#Mean error per data point (half-length of error bar)\n#is %.2f gauss.\n", sigmasq);
}
